require 'twitter_cldr'

module Mongoid
  module Contextual
    class Mongo
      def sort_with_collation(values = nil, &block)
        if klass.respond_to?(:collation_aliases)
          values = replace_collation_aliases(values)
        end

        sort_without_collation(values, &block)
      end

      alias_method_chain :sort, :collation

      private

      def replace_collation_aliases(values)
        case values
          when Hash
            values.dup.tap do |values_copy|
              values_copy.keys.each do |key|
                if is_collated?(key)
                  values_copy[collated_alias(key)] = values[key]
                  values_copy.delete(key)
                end
              end
            end
          else
            Array(values).dup.tap do |values_copy|
              values_copy.each do |key|
                if key.is_a?(Origin::Key)
                  if is_collated?(key.name)
                    values_copy.delete(key)

                    values_copy <<
                      rebuild_origin_key(key, collated_alias(key.name))
                  end
                end
              end
            end
        end
      end

      def is_collated?(key)
        klass.collation_aliases.has_key?(key.to_s)
      end

      def collated_alias(key)
        klass.collation_aliases[key.to_s]
      end

      def rebuild_origin_key(key, name)
        ::Origin::Key.new(name, key.strategy, key.operator, key.expanded,
                          &key.block)
      end
    end
  end

  module Collation
    extend ActiveSupport::Concern

    included do
      class_attribute :collation_aliases
      self.collation_aliases = {}
    end

    module ClassMethods
      def collation(field, options = {})
        raise ':locale option is required' unless options[:locale]

        options[:suffix] ||= :collated

        field_collated = "#{field}_#{options[:suffix]}"
        field field_collated, type: String

        self.collation_aliases[field.to_s] = field_collated

        collator = TwitterCldr::Collation::Collator.new(options[:locale])

        before_save do
          sort_key = collator.get_sort_key(send(field))
          send("#{field_collated}=", sort_key.map { |i| i.to_s(16) }.join)
        end
      end
    end
  end
end
